from base import BaseScanner
import platform

import pyinotify

class EventHandler(pyinotify.ProcessEvent):
    def __init__(self, scanner):
        self._scanner = scanner

    def process_IN_CREATE(self, event):
        if self._scanner.is_valid_type(event.pathname):
            self._scanner.trigger_created(event.pathname)

    def process_IN_DELETE(self, event):
        if self._scanner.is_valid_type(event.pathname):
            self._scanner.trigger_deleted(event.pathname)

    def process_IN_MODIFY(self, event):
        #self._process('modified', event.pathname)
        if self._scanner.is_valid_type(event.pathname):
            self._scanner.trigger_modified(event.pathname)


class PyINotifyScanner(BaseScanner):
    """
    Scanner that uses pyinotify (alias, inotify) for notification events.
    """
    def __init__(self, *args, **kwargs):
        super(PyINotifyScanner, self).__init__(*args, **kwargs)
        self.log("Library of choice: pyinotify")
        self._watcher = pyinotify.WatchManager()
        self._notifier = self._generate_notifier()

    def __deinit__(self):
        if self._notifier is not None:
            self._notifier.stop()

    def _generate_notifier(self):
        handler = EventHandler(self)

        notifier = pyinotify.Notifier(self._watcher, handler)
        mask = pyinotify.IN_DELETE | pyinotify.IN_CREATE | pyinotify.IN_MODIFY
        for path in self.paths:
            self._watcher.add_watch(path, mask, rec=True, auto_add=True,
                                    exclude_filter=self.is_valid_type)
        
        return notifier

    def loop(self, sleep_time=None, callback=None):
        self.trigger_init()
        try:
            self._notifier.loop(callback)
        except KeyboardInterrupt:
            self._notifier.stop()
            raise

    def step(self):
        self._notifier.process_events()
        if self._notifier.check_events(timeout=1000):
            self.read_events()
        self._notifier.stop()

    def stop(self):
        self._notifier.stop()



from modules_restore_point import ModulesRestorePoint
from functools import wraps
from termstyle import bg_red, bg_green, white
import platform
import os
import sys
import scent_picker

__all__ = ['Sniffer']

# for debugging
def echo(text):
    def wrapped(filepath):
        print text % {'file': filepath}
    return wrapped

class Sniffer(object):
    """
    Handles the execution of the sniffer. The interface that main.run expects is:

    ``set_up(test_args, clear, debug)``
    
      ``test_args`` The arguments to pass to the test runner.
      ``clear``     Boolean. Set to True if we should clear console before running
                    the tests.
      ``debug``     Boolean. Set to True if we want to print debugging information.

    ``observe_scanner(scanner)``

      ``scanner``   The scanner instance to hook events into. By default, ``self._run`` is
                    attached, which then calls self.run(). The run method should return
                    True on passing and False on failure.
    """
    def __init__(self):
        self.modules = ModulesRestorePoint()
        self._scanners = []
        self.pass_colors = {'fg': white, 'bg': bg_green}
        self.fail_colors = {'fg': white, 'bg': bg_red}
        self.set_up()
        
    def set_up(self, test_args=(), clear=True, debug=False):
        """
        Sets properties right before calling run.

          ``test_args`` The arguments to pass to the test runner.
          ``clear``     Boolean. Set to True if we should clear console before running
                        the tests.
          ``debug``     Boolean. Set to True if we want to print debugging information.
        """
        self.test_args = test_args
        self.debug, self.clear = debug, clear
        
    def absorb_args(self, func):
        """
        Calls a function without any arguments. The returned caller function
        accepts any arguments (and throws them away).
        """
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func()
        return wrapper
    
    def observe_scanner(self, scanner):
        """
        Hooks into multiple events of a scanner.
        """
        scanner.observe(scanner.ALL_EVENTS, self.absorb_args(self.modules.restore))
        if self.clear:
            scanner.observe(scanner.ALL_EVENTS, self.absorb_args(self.clear_on_run))
        scanner.observe(scanner.ALL_EVENTS, self.absorb_args(self._run))
        if self.debug:
            scanner.observe('created',  echo("callback - created  %(file)s"))
            scanner.observe('modified', echo("callback - changed  %(file)s"))
            scanner.observe('deleted',  echo("callback - deleted  %(file)s"))
        self._scanners.append(scanner)

    def clear_on_run(self, prefix="Running Tests:"):
        """Clears console before running the tests."""
        if platform.system() == 'Windows':
            os.system('cls')
        else:
            os.system('clear')
        if prefix:
            print prefix

    def _stop(self):
        """Calls stop() to all scanner in an attempt to quit."""
        for scanner in self._scanners:
            scanner.stop()

    def _run(self):
        """Calls self.run() and wraps for errors."""
        try:
            if self.run():
                print self.pass_colors['bg'](self.pass_colors['fg']("In good standing"))
            else:
                print self.fail_colors['bg'](self.fail_colors['fg']("Failed - Back to work!"))
        except StandardError:
            import traceback
            traceback.print_exc()
            self._stop()
            raise
        except Exception:
            self._stop()
            raise
        return True

    def run(self):
        """
        Runs the unit test framework. Can be overridden to run anything.
        Returns True on passing and False on failure.
        """
        # import here instead of on top:
        # - To force a module reload to resolve this bug until Nose fixes it:
        #   http://github.com/gfxmonk/autonose/issues#issue/13
        try:
            import nose
            arguments = [sys.argv[0]] + list(self.test_args)
            return nose.run(argv=arguments)
        except ImportError:
            print
            print "*** Nose library missing. Please install it. ***"
            print
            raise

class ScentSniffer(Sniffer):
    """Runs arbitrary python code in the cwd's scent.py file."""
    def __init__(self, cwd=None, scent="scent.py"):
        self.cwd = cwd or os.getcwd()
        self.scent = scent_picker.exec_from_dir(self.cwd, scent)
        super(ScentSniffer, self).__init__()
        self.update_from_scent()
        
    def update_from_scent(self):
        if self.scent:
            self.pass_colors['fg'] = self.scent.fg_pass
            self.pass_colors['bg'] = self.scent.bg_pass
            self.fail_colors['fg'] = self.scent.fg_fail
            self.fail_colors['bg'] = self.scent.bg_fail
        
    def refresh_scent(self, filepath):
        if self.scent and filepath == self.scent.filename:
            print "Reloaded Scent:", filepath
            for s in self._scanners:
                self.unobserve_scanner(s)
            self.scent = self.scent.reload()
            self.update_from_scent()
            for s in self._scanners:
                self.scent_observe_scanner(s)
                
    def unobserve_scanner(self, scanner):
        for v in self.scent.validators:
            if self.debug:
                print "Removed", repr(v)
            scanner.remove_validator(v)
            
    def scent_observe_scanner(self, scanner):
        if self.scent:
            for v in self.scent.validators:
                if self.debug:
                    print "Added", repr(v)
                scanner.add_validator(v)
    
    def observe_scanner(self, scanner):            
        scanner.observe('created', self.refresh_scent)
        scanner.observe('modified', self.refresh_scent)
        self.scent_observe_scanner(scanner)
        return super(ScentSniffer, self).observe_scanner(scanner)
        
    def clear_on_run(self):
        super(ScentSniffer, self).clear_on_run(None)
        
    def run(self):
        """
        Runs the CWD's scent file.
        """
        if not self.scent or len(self.scent.runners) == 0:
            print "Did not find 'scent.py', running nose:"
            return super(ScentSniffer, self).run()
        else:
            print "Using scent:"
            arguments = [sys.argv[0]] + list(self.test_args)
            return self.scent.run(arguments)
        return True



g = globals().copy()
import os, sys, termstyle

class ScentModule(object):
    def __init__(self, mod, filename):
        self.mod = mod
        self.filename = filename
        self.validators = []
        self.runners = []
        for name in dir(self.mod):
            obj = getattr(self.mod, name)
            type = getattr(obj, 'scent_api_type', None)
            if type == 'runnable':
                self.runners.append(obj)
            elif type == 'file_validator':
                self.validators.append(obj)
        self.runners = tuple(self.runners)
        self.validators = tuple(self.validators)
        print self.validators
        
    def reload(self):
        try:
            return load_file(self.filename)
        except Exception:
            import traceback
            traceback.print_exc()
            print
            print "Still using previously valid Scent."
            return self
    
    def run(self, args):
        try:
            for r in self.runners:
                if not r(*args):
                    return False
            return True
        except Exception:
            import traceback
            traceback.print_exc()
            print
            return False
        
    @property
    def fg_pass(self):
        return getattr(self.mod, 'pass_fg_color', termstyle.black)
    @property
    def bg_pass(self):
        return getattr(self.mod, 'pass_bg_color', termstyle.bg_green)
    
    @property
    def fg_fail(self):
        return getattr(self.mod, 'fail_fg_color', termstyle.white)
    @property
    def bg_fail(self):
        return getattr(self.mod, 'fail_bg_color', termstyle.bg_red)

def load_file(filename):
    "Runs the given scent.py file."
    mod_name = '.'.join(os.path.basename(filename).split('.')[:-1])
    mod_path = os.path.dirname(filename)
    
    global_vars = globals()
    if mod_name in sys.modules.keys():
        del sys.modules[mod_name]
    if mod_path not in set(sys.modules.keys()):
        sys.path.insert(0, mod_path)
    return ScentModule(__import__(mod_name, g, g), filename)
    
def exec_from_dir(dirname=None, scent="scent.py"):
    """Runs the scent.py file from the given directory (cwd if None given).
    
    Returns module if loaded a scent, None otherwise.
    """
    if dirname is None:
        dirname = os.getcwd()
    files = os.listdir(dirname)
    
    if scent not in files:
        return None
    
    return load_file(os.path.join(dirname, scent))



"""
Sniffer - An automatic test runner for Nose
==========

This is a simple program that runs nose anytime the current working
directory (or its subdirectories) have changed (added, modified, or
delete files).

*Note:* By default, sniffer polls the directory tree to determine what changed.
This method was used because of its:

 - simplicity: It's pretty easy to write.
 - cross-platform: Unlike third-party libs, which only work on one OS.
 - standard: Uses only standard modules.

Alternatively, third party modules can be installed to increase performance.
The library to install is dependent on your operating system:

 - Linux: install pyinotify
 - Windows: install pywin32
 - OSX: install MacFSEvents

"""
from metadata import *
from scanner import Scanner
from runner import Sniffer
from main import main, run

__all__ = ['Scanner', 'Sniffer', 'main', 'run']



__all__ = [
    '__author__', '__author_email__', '__copyright__', '__credits__',
    '__license__', '__version__'
]


__author__ = "Jeff Hui"
__author_email__ = 'contrib@jeffhui.net'
__copyright__ = "Copyright 2010, Jeff Hui"
__credits__ = ["Jeff Hui"]

__license__ = "MIT"
__version__ = "0.2.3"



from setuptools import setup, find_packages

# will fail to install via pip/easy_install
#from sniffer.metadata import __version__, __author__, __author_email__

__version__, __author__, __author_email__ = "0.2.3", "Jeff Hui", "contrib@jeffhui.net"

setup(
    name='sniffer',
    version=__version__,
    description='An automatic test runner. Supports nose out of the box.',
    long_description=open('README.rst').read(),
    author=__author__,
    author_email=__author_email__,
    url='http://github.com/jeffh/sniffer/',
    install_requires=[
        'python-termstyle',
        'nose',
    ],
    entry_points = {
        'console_scripts': ['sniffer = sniffer:main'],
    },
    packages=find_packages(),
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Testing',
        'Topic :: Software Development :: Libraries',
    ],
)



"""
Provides the scanner class which deals with monitoring and notifying changing
in the file system.

It provides a polling technique which is an OS-independent and uses no third-party
libraries at the cost of performance. The polling technique constantly walks through
the directory tree to see which files changed, calling os.stat on the files.
"""
from base import PollingScanner

__all__ = ['Scanner']

Scanner = PollingScanner

# A messy way to import library-specific classes.
# I should really make a factory class or something, but I'm lazy.
# Plus, factories remind me a lot of java...
def _import(module, cls):
    global Scanner
    try:
        cls = str(cls)
        mod = __import__(str(module), globals(), locals(), [cls], -1)
        Scanner = getattr(mod, cls)
    except ImportError:
        pass

_import('pywin_scanner', 'PyWinScanner') # windows
_import('fsevents_scanner', 'FSEventsScanner') # osx
_import('pyinotify_scanner', 'PyINotifyScanner') # linux
